/* tcldom-libxml.c --
 *
 *	A Tcl wrapper for libxml's node tree API,
 *	conformant to the TclDOM API.
 *
 * Copyright (c) 2001,2002 Zveno Pty Ltd
 * http://www.zveno.com/
 *
 * Zveno Pty Ltd makes this software and associated documentation
 * available free of charge for any purpose.  You may make copies
 * of the software but you must include all of this notice on any copy.
 *
 * Zveno Pty Ltd does not warrant that this software is error free
 * or fit for any purpose.  Zveno Pty Ltd disclaims any liability for
 * all claims, expenses, losses, damages and costs any user may incur
 * as a result of using, copying or modifying the software.
 *
 * $Id: tcldom-libxml2.c,v 1.9 2002/02/27 04:28:39 balls Exp $
 */

#include "tcldom.h"
#include <libxml/xpath.h>
#include <libxml/xmlIO.h>

#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT

/*
 * Data structures for internal functions.
 */

typedef struct ParserClientData {
  Tcl_Interp *interp;
  Tcl_Obj *externalentityloader;	/* Callback for loading an external entity */
} ParserClientData;

/*
 * Prototypes for procedures defined later in this file:
 */

Tcl_FreeInternalRepProc	TclDOM_DocFree;
Tcl_DupInternalRepProc	TclDOM_DocDup;
Tcl_UpdateStringProc	TclDOM_DocUpdate;
Tcl_SetFromAnyProc	TclDOM_DocSetFromAny;

Tcl_FreeInternalRepProc	TclDOM_NodeFree;
Tcl_DupInternalRepProc	TclDOM_NodeDup;
Tcl_UpdateStringProc	TclDOM_NodeUpdate;
Tcl_SetFromAnyProc	TclDOM_NodeSetFromAny;

/*
 * Forward declarations for private functions.
 */

static void TclDOMGenericError _ANSI_ARGS_((void *ctx, const char *msg, ...));
static Tcl_Obj * TclDOMNewDoc _ANSI_ARGS_((Tcl_Interp *interp));

Tcl_Obj * TclDOM_CreateObjFromNode _ANSI_ARGS_((xmlNodePtr nodePtr));
Tcl_Obj * TclDOM_CreateObjFromDoc  _ANSI_ARGS_((xmlDocPtr  docPtr));


/*
 * Forward declarations of commands
 */

static int TclDOMDOMImplementationCommand _ANSI_ARGS_((ClientData dummy,
						       Tcl_Interp *interp,
						       int objc,
						       Tcl_Obj *CONST objv[]));
static int TclDOMDocumentCommand _ANSI_ARGS_((ClientData dummy,
					      Tcl_Interp *interp,
					      int objc,
					      Tcl_Obj *CONST objv[]));
static int TclDOMNodeCommand _ANSI_ARGS_((ClientData dummy,
					  Tcl_Interp *interp,
					  int objc,
					  Tcl_Obj *CONST objv[]));
static int TclDOMElementCommand _ANSI_ARGS_((ClientData dummy,
					     Tcl_Interp *interp,
					     int objc,
					     Tcl_Obj *CONST objv[]));
static int TclDOMEventCommand _ANSI_ARGS_((ClientData dummy,
					   Tcl_Interp *interp,
					   int objc,
					   Tcl_Obj *CONST objv[]));

/*
 * Functions that implement the TclDOM_Implementation interface
 */

static int TclDOMCreateCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));
static int TclDOMParseCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));
static int TclDOMSerializeCommand _ANSI_ARGS_((ClientData dummy,
					    Tcl_Interp *interp,
					    int objc,
					    Tcl_Obj *CONST objv[]));

/*
 * libxml2 callbacks
 */

static xmlParserInputPtr TclDOMExternalEntityLoader _ANSI_ARGS_((const char *URL,
								 const char *ID,
								 xmlParserCtxtPtr ctxt));

/*
 * Private libxml2 functions
 */

EXTERN xmlParserCtxtPtr xmlCreateMemoryParserCtxt _ANSI_ARGS_((const char *buffer,
							      int size));

Tcl_ObjType TclDOM_DocObjType = {
  "libxml2-doc",
  TclDOM_DocFree,
  TclDOM_DocDup,
  TclDOM_DocUpdate,
  TclDOM_DocSetFromAny
};

Tcl_ObjType TclDOM_NodeObjType = {
  "libxml2-node",
  TclDOM_NodeFree,
  TclDOM_NodeDup,
  TclDOM_NodeUpdate,
  TclDOM_NodeSetFromAny
};

/*
 * Hash table for mapping string rep of documents to doc structure.
 */

static Tcl_HashTable documents;
static int docCntr = 0;

/*
 * default loader is overridden while parsing
 */

static xmlExternalEntityLoader defaultLoader;

/*
 * Switch tables
 */

static char *DOMImplementationCommandMethods[] = {
  "hasFeature",
  "createDocument",
  "create",
  "createDocumentType",
  "createNode",
  "destroy",
  "isNode",
  "parse",
  "selectNode",
  "serialize",
  "trim",
  (char *) NULL
};
enum DOMImplementationCommandMethods {
  TCLDOM_IMPL_HASFEATURE,
  TCLDOM_IMPL_CREATEDOCUMENT,
  TCLDOM_IMPL_CREATE,
  TCLDOM_IMPL_CREATEDOCUMENTYPE,
  TCLDOM_IMPL_CREATENODE,
  TCLDOM_IMPL_DESTROY,
  TCLDOM_IMPL_ISNODE,
  TCLDOM_IMPL_PARSE,
  TCLDOM_IMPL_SELECTNODE,
  TCLDOM_IMPL_SERIALIZE,
  TCLDOM_IMPL_TRIM
};
static char *DocumentCommandMethods[] = {
  "cget",
  "configure",
  "createElement",
  "createDocumentFragment",
  "createTextNode",
  "createComment",
  "createCDATASection",
  "createProcessingInstruction",
  "createAttribute",
  "createEntity",
  "createEntityReference",
  "createDocTypeDecl",
  "importNode",
  "createElementNS",
  "createAttributeNS",
  "getElementsByTagNameNS",
  "getElementsById",
  "createEvent",
  "getElementsByTagName",
  (char *) NULL
};
enum DocumentCommandMethods {
  TCLDOM_DOCUMENT_CGET,
  TCLDOM_DOCUMENT_CONFIGURE,
  TCLDOM_DOCUMENT_CREATEELEMENT,
  TCLDOM_DOCUMENT_CREATEDOCUMENTFRAGMENT,
  TCLDOM_DOCUMENT_CREATETEXTNODE,
  TCLDOM_DOCUMENT_CREATECOMMENT,
  TCLDOM_DOCUMENT_CREATECDATASECTION,
  TCLDOM_DOCUMENT_CREATEPI,
  TCLDOM_DOCUMENT_CREATEATTRIBUTE,
  TCLDOM_DOCUMENT_CREATEENTITY,
  TCLDOM_DOCUMENT_CREATEENTITYREFERENCE,
  TCLDOM_DOCUMENT_CREATEDOCTYPEDECL,
  TCLDOM_DOCUMENT_IMPORTNODE,
  TCLDOM_DOCUMENT_CREATEELEMENTNS,
  TCLDOM_DOCUMENT_CREATEATTRIBUTENS,
  TCLDOM_DOCUMENT_GETELEMENTSBYTAGNAMENS,
  TCLDOM_DOCUMENT_GETELEMENTSBYID,
  TCLDOM_DOCUMENT_CREATEEVENT,
  TCLDOM_DOCUMENT_GETELEMENTSBYTAGNAME
};
static char *DocumentCommandOptions[] = {
  "-doctype",
  "-implementation",
  "-documentElement",
  (char *) NULL
};
enum DocumentCommandOptions {
  TCLDOM_DOCUMENT_DOCTYPE,
  TCLDOM_DOCUMENT_IMPLEMENTATION,
  TCLDOM_DOCUMENT_DOCELEMENT
};
static char *NodeCommandMethods[] = {
  "cget",
  "configure",
  "insertBefore",
  "replaceChild",
  "removeChild",
  "appendChild",
  "hasChildNodes",
  "cloneNode",
  "children",
  "parent",
  "path",
  "createNode",
  "selectNode",
  "stringValue",
  "addEventListener",
  "removeEventListener",
  "dispatchEvent",
  (char *) NULL
};
enum NodeCommandMethods {
  TCLDOM_NODE_CGET,
  TCLDOM_NODE_CONFIGURE,
  TCLDOM_NODE_INSERTBEFORE,
  TCLDOM_NODE_REPLACECHILD,
  TCLDOM_NODE_REMOVECHILD,
  TCLDOM_NODE_APPENDCHILD,
  TCLDOM_NODE_HASCHILDNODES,
  TCLDOM_NODE_CLONENODE,
  TCLDOM_NODE_CHILDREN,
  TCLDOM_NODE_PARENT,
  TCLDOM_NODE_PATH,
  TCLDOM_NODE_CREATENODE,
  TCLDOM_NODE_SELECTNODE,
  TCLDOM_NODE_STRINGVALUE,
  TCLDOM_NODE_ADDEVENTLISTENER,
  TCLDOM_NODE_REMOVEEVENTLISTENER,
  TCLDOM_NODE_DISPATCHEVENT
};
static char *NodeCommandOptions[] = {
  "-nodeType",
  "-parentNode",
  "-childNodes",
  "-firstChild",
  "-lastChild",
  "-previousSibling",
  "-nextSibling",
  "-attributes",
  "-namespaceURI",
  "-prefix",
  "-localName",
  "-nodeValue",
  "-cdatasection",
  "-nodeName",
  (char *) NULL
};
enum NodeCommandOptions {
  TCLDOM_NODE_NODETYPE,
  TCLDOM_NODE_PARENTNODE,
  TCLDOM_NODE_CHILDNODES,
  TCLDOM_NODE_FIRSTCHILD,
  TCLDOM_NODE_LASTCHILD,
  TCLDOM_NODE_PREVIOUSSIBLING,
  TCLDOM_NODE_NEXTSIBLING,
  TCLDOM_NODE_ATTRIBUTES,
  TCLDOM_NODE_NAMESPACEURI,
  TCLDOM_NODE_PREFIX,
  TCLDOM_NODE_LOCALNAME,
  TCLDOM_NODE_NODEVALUE,
  TCLDOM_NODE_CDATASECTION,
  TCLDOM_NODE_NODENAME
};
static char *ElementCommandMethods[] = {
  "cget",
  (char *) NULL
};
enum ElementCommandMethods {
  TCLDOM_ELEMENT_CGET
};
static char *ParseCommandOptions[] = {
  "-baseuri",
  "-externalentitycommand",
  (char *) NULL
};
enum ParseCommandOptions {
  TCLDOM_PARSE_BASEURI,
  TCLDOM_PARSE_EXTERNALENTITYCOMMAND
};

/*
 * Error context for passing error result back to caller.
 */

typedef struct GenericError_Info {
  Tcl_Interp *interp;
  int code;
  Tcl_Obj *msg;
} GenericError_Info;

static GenericError_Info errorInfo = {
  NULL,
  TCL_OK,
  NULL
};

/*
 * Default values
 */

EXTERN int xmlLoadExtDtdDefaultValue;


/*
 *----------------------------------------------------------------------------
 *
 * Tcldomxml_Init --
 *
 *  Initialisation routine for loadable module
 *
 * Results:
 *  None.
 *
 * Side effects:
 *  Creates commands in the interpreter,
 *
 *----------------------------------------------------------------------------
 */

int
Tcldomxml_Init (interp)
     Tcl_Interp *interp;	/* Interpreter to initialise */
{
  TclDOM_Implementation *implPtr;

  Tcl_InitStubs(interp, "8.1", 0);

  /* Configure the libxml2 parser */
  xmlInitMemory();
  xmlSubstituteEntitiesDefault(1);

  /*
   * 6 will load external entities.
   * 0 will not.
   * TODO: provide configuration option for setting this value.
   */
  xmlLoadExtDtdDefaultValue = 6;

  xmlSetGenericErrorFunc((void *) &errorInfo, TclDOMGenericError);

  Tcl_RegisterObjType(&TclDOM_DocObjType);
  Tcl_RegisterObjType(&TclDOM_NodeObjType);

  /*
   * Register this DOM implementation with the TclDOM
   * generic layer.  We must do this for each separate
   * Tcl object type.
   */

  implPtr = (TclDOM_Implementation *) Tcl_Alloc(sizeof(TclDOM_Implementation));
  implPtr->name = Tcl_NewStringObj("libxml2-doc", -1);
  implPtr->type = &TclDOM_DocObjType;
  implPtr->create = TclDOMCreateCommand;
  implPtr->parse = TclDOMParseCommand;
  implPtr->serialize = TclDOMSerializeCommand;
  implPtr->document = TclDOMDocumentCommand;
  implPtr->documentfragment = NULL;
  implPtr->node = NULL;
  implPtr->element = NULL;
  TclDOM_Register(interp, implPtr);

  implPtr = (TclDOM_Implementation *) Tcl_Alloc(sizeof(TclDOM_Implementation));
  implPtr->name = Tcl_NewStringObj("libxml2-node", -1);
  implPtr->type = &TclDOM_NodeObjType;
  implPtr->create = TclDOMCreateCommand;
  implPtr->parse = TclDOMParseCommand;
  implPtr->serialize = TclDOMSerializeCommand;
  implPtr->document = TclDOMDocumentCommand;
  implPtr->documentfragment = NULL;
  implPtr->node = TclDOMNodeCommand;
  implPtr->element = TclDOMElementCommand;
  TclDOM_Register(interp, implPtr);

  Tcl_CreateObjCommand(interp, "dom::libxml2::DOMImplementation", TclDOMDOMImplementationCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::document", TclDOMDocumentCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::node", TclDOMNodeCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::create", TclDOMCreateCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::parse", TclDOMParseCommand, NULL, NULL);
  Tcl_CreateObjCommand(interp, "dom::libxml2::serialize", TclDOMSerializeCommand, NULL, NULL);
/*
  Tcl_CreateObjCommand(interp, "dom::libxml2::element", TclDOMElementCommand, NULL, NULL);
*/

  Tcl_InitHashTable(&documents, TCL_STRING_KEYS);

  Tcl_PkgProvide(interp, "dom::libxml2", "1.0");

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMCreateCommand --
 *
 *  Implements dom::libxml2::create command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMCreateCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  if (objc != 1) {
    Tcl_WrongNumArgs(interp, 1, objv, "");
    return TCL_ERROR;
  }

  Tcl_SetObjResult(interp, TclDOMNewDoc(interp));
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMParseCommand --
 *
 *  Implements dom::libxml2::parse command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMParseCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  char *buf;
  int len, new, option;
  xmlParserCtxtPtr ctxt;
  xmlDocPtr docPtr;
  Tcl_Obj *objPtr;
  Tcl_HashEntry *entryPtr;
  Tcl_Obj *baseuriPtr = NULL;
  Tcl_Obj *extentity = NULL;
  ParserClientData *clientData;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "xml ?args ...?");
    return TCL_ERROR;
  }

  buf = Tcl_GetStringFromObj(objv[1], &len);
  if (buf == NULL) {
    Tcl_SetResult(interp, "unable to get document to parse", NULL);
    return TCL_ERROR;
  }

  /*
   * Process options
   */

  objc -= 2;
  objv += 2;
  while (objc) {
    if (objc == 1) {
      Tcl_Obj *msgPtr;

      msgPtr = Tcl_NewStringObj("missing value for configuration option \"", -1);
      Tcl_AppendObjToObj(msgPtr, objv[0]);
      Tcl_AppendStringsToObj(msgPtr, "\"", (char *) NULL);
      Tcl_SetObjResult(interp, msgPtr);
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[0], ParseCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum ParseCommandOptions) option) {
    case TCLDOM_PARSE_BASEURI:
      baseuriPtr = objv[1];
      break;

    case TCLDOM_PARSE_EXTERNALENTITYCOMMAND:
      extentity = objv[1];
      break;
    }

    objc -= 2;
    objv += 2;
  }

  /*
   * This does the real work
   */

  xmlInitParser();
  ctxt = xmlCreateMemoryParserCtxt(buf, len);
  if (ctxt == NULL) {
    /* Out of memory - we're in big trouble... */
    Tcl_SetResult(interp, "unable to allocate parser context", NULL);
    return TCL_ERROR;
  }

  /*
   * Use the _private field to store TclDOM data
   */

  clientData = (ParserClientData *) Tcl_Alloc(sizeof(ParserClientData));
  ctxt->_private = (char *) clientData;
  clientData->interp = interp;

  if (baseuriPtr) {
    ctxt->input->filename = Tcl_GetStringFromObj(baseuriPtr, NULL);
  }

  defaultLoader = xmlGetExternalEntityLoader();
  if (extentity) {
    clientData->externalentityloader = extentity;
    xmlSetExternalEntityLoader(TclDOMExternalEntityLoader);
  }

  /* docPtr = xmlParseMemory(buf, len); */
  if (xmlParseDocument(ctxt) != 0) {
    docPtr = NULL;
  } else {
    docPtr = ctxt->myDoc;
  }

  xmlSetExternalEntityLoader(defaultLoader);
  xmlFreeParserCtxt(ctxt);
  xmlCleanupParser();

  if (docPtr == NULL) {
    Tcl_SetResult(interp, "unable to parse document", NULL);
    return TCL_ERROR;
  }

  /*
   * Wrap the document pointer in a Tcl object
   */

  buf = Tcl_Alloc(20);
  sprintf(buf, "doc%d", docCntr++);

  entryPtr = Tcl_CreateHashEntry(&documents, buf, &new);
  if (!new) {
    Tcl_Free(buf);
    Tcl_SetResult(interp, "internal error: previously allocated token", NULL);
    return TCL_ERROR;
  }
  Tcl_SetHashValue(entryPtr, (void *) docPtr);

  objPtr = Tcl_NewObj();
  objPtr->internalRep.otherValuePtr = (VOID *) docPtr;
  objPtr->typePtr = &TclDOM_DocObjType;
  objPtr->bytes = Tcl_Alloc(20);
  strcpy(objPtr->bytes, buf);
  objPtr->length = strlen(buf);

  Tcl_Free(buf);

  Tcl_SetObjResult(interp, objPtr);

  return TCL_OK;
}

/*
 * Interpose on resolving external entity references.
 *
 * The return code of the script evaluation determines
 * the behaviour:
 * TCL_OK - return value is to be used as the entity data
 * TCL_CONTINUE - use the default libxml2 entity loader
 * TCL_BREAK - forget about this entity
 * TCL_ERROR - background error
 */

static xmlParserInputPtr
TclDOMExternalEntityLoader(url, id, ctxt)
     const char *url;
     const char *id;
     xmlParserCtxtPtr ctxt;
{
  ParserClientData *clientData = (ParserClientData *) ctxt->_private;
  Tcl_Obj *cmdPtr;

  if (clientData) {
    cmdPtr = Tcl_DuplicateObj(clientData->externalentityloader);
  } else {
    return NULL;
  }

  if (url) {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewStringObj(url, -1));
  } else {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewListObj(0, NULL));
  }
  if (id) {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewStringObj(id, -1));
  } else {
    Tcl_ListObjAppendElement(clientData->interp, cmdPtr, Tcl_NewListObj(0, NULL));
  }

  switch (Tcl_GlobalEvalObj(clientData->interp, cmdPtr)) {

  case TCL_OK:
    /* TODO: Use the return value as the input */
    return (defaultLoader)(url, id, ctxt);

  case TCL_CONTINUE:
    /* Use the default libxml2 loader */
    return (defaultLoader)(url, id, ctxt);

  case TCL_BREAK:
    /* Do not load external entity, but no error */
    return NULL;

  default:
    Tcl_BackgroundError(clientData->interp);
    return NULL;
  }
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMSerializeCommand --
 *
 *  Implements dom::libxml2::serialize command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMSerializeCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  xmlDocPtr docPtr;
  xmlNodePtr nodePtr;
  xmlChar *result;
  int len;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "node ?args ...?");
    return TCL_ERROR;
  }

  if (TclDOM_GetDocFromObj(interp, objv[1], &docPtr) != TCL_OK) {
    if (TclDOM_GetNodeFromObj(interp, objv[1], &nodePtr) == TCL_OK) {
      /* Serialize just the node */
      Tcl_SetResult(interp, "not yet implemented - serialize whole document", NULL);
      return TCL_ERROR;
    } else {
      Tcl_SetResult(interp, "not a libxml2 node", NULL);
      return TCL_ERROR;
    }
  }

  xmlDocDumpMemory(docPtr, &result, &len);
  Tcl_SetObjResult(interp, Tcl_NewStringObj(result, len));
  free(result);

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMDOMImplementationCommand --
 *
 *  Implements dom::libxml2::DOMImplementation command
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMDOMImplementationCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], DOMImplementationCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  switch ((enum DOMImplementationCommandMethods) method) {
  case TCLDOM_IMPL_CREATE:
    if (objc == 2) {
      return TclDOMCreateCommand(dummy, interp, 0, NULL);
    } else if (objc == 3) {
      Tcl_Obj *objPtr;
      xmlDocPtr docPtr;
      xmlNodePtr nodePtr;

      if (TclDOMCreateCommand(dummy, interp, 0, NULL) != TCL_OK) {
	return TCL_ERROR;
      }
      objPtr = Tcl_GetObjResult(interp);
      TclDOM_GetDocFromObj(interp, objPtr, &docPtr);
      nodePtr = xmlNewDocNode(docPtr, NULL, Tcl_GetStringFromObj(objv[2], NULL), NULL);
      if (nodePtr == NULL) {
	Tcl_SetResult(interp, "unable to create document element", NULL);
	Tcl_DecrRefCount(objPtr);
	return TCL_ERROR;
      }

      Tcl_SetObjResult(interp, objPtr);
    } else {
      Tcl_WrongNumArgs(interp, 1, objv, "create ?doc?");
      return TCL_ERROR;
    }

    break;

  case TCLDOM_IMPL_PARSE:
    return TclDOMParseCommand(dummy, interp, objc, objv);

  case TCLDOM_IMPL_SERIALIZE:
    return TclDOMSerializeCommand(dummy, interp, objc, objv);

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMDocumentCommand --
 *
 *  Implements dom::libxml2::document command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMDocumentCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option;
  xmlDocPtr docPtr = NULL;
  xmlNodePtr nodePtr = NULL;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "method token ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], DocumentCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetDocFromObj(interp, objv[2], &docPtr) != TCL_OK) {
    docPtr = NULL;
    if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
      return TCL_ERROR;
    }
  }

  switch ((enum DocumentCommandMethods) method) {

  case TCLDOM_DOCUMENT_CGET:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "cget option");
      return TCL_ERROR;
    }

    if (!docPtr) {
      Tcl_SetResult(interp, "not a document", NULL);
    }

    if (Tcl_GetIndexFromObj(interp, objv[3], DocumentCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum DocumentCommandOptions) option) {

    case TCLDOM_DOCUMENT_DOCTYPE:
    case TCLDOM_DOCUMENT_IMPLEMENTATION:
      Tcl_SetResult(interp, "not yet implemented", NULL);
      return TCL_ERROR;

    case TCLDOM_DOCUMENT_DOCELEMENT:

      nodePtr = xmlDocGetRootElement(docPtr);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(nodePtr));

      break;

    default:
      Tcl_SetResult(interp, "unknown option", NULL);
      return TCL_ERROR;
    }

    break;

  case TCLDOM_DOCUMENT_CONFIGURE:

    if (objc == 4) {
      Tcl_Obj *newobjv[4];

      newobjv[0] = objv[0];
      newobjv[1] = Tcl_NewStringObj("cget", -1);
      newobjv[2] = objv[2];
      newobjv[3] = objv[3];

      if (TclDOMDocumentCommand(NULL, interp, 4, newobjv) != TCL_OK) {
	return TCL_ERROR;
      }

    } else {
      Tcl_SetResult(interp, "options are read-only", NULL);
    }

    break;

  case TCLDOM_DOCUMENT_CREATEELEMENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token name");
      return TCL_ERROR;
    }

    if (docPtr) {
      xmlNodePtr new = xmlNewDocNode(docPtr, NULL, Tcl_GetStringFromObj(objv[3], NULL), NULL);
      xmlNodePtr old = xmlDocSetRootElement(docPtr, new);
      if (old) {
	xmlDocSetRootElement(docPtr, old);
	xmlFreeNode(new);
	Tcl_SetResult(interp, "document element already exists", NULL);
	return TCL_ERROR;
      }
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(new));
    } else {
      xmlNodePtr new = xmlNewChild(nodePtr, NULL, Tcl_GetStringFromObj(objv[3], NULL), NULL);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(new));
    }

    break;

  case TCLDOM_DOCUMENT_CREATEDOCUMENTFRAGMENT:

    if (objc != 3) {
      Tcl_WrongNumArgs(interp, 2, objv, "token");
      return TCL_ERROR;
    }

    if (docPtr) {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocFragment(docPtr)));
    } else {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocFragment(nodePtr->doc)));
    }

    break;

  case TCLDOM_DOCUMENT_CREATETEXTNODE:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token text");
      return TCL_ERROR;
    }

    if (docPtr) {
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocTextLen(docPtr, content, len)));
    } else {
      xmlNodePtr textnode;
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      textnode = xmlNewTextLen(content, len);
      xmlAddChild(nodePtr, textnode);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(textnode));
    }

    break;

  case TCLDOM_DOCUMENT_CREATECOMMENT:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token data");
      return TCL_ERROR;
    }

    if (docPtr) {
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewComment(Tcl_GetStringFromObj(objv[3], NULL))));
    } else {
      xmlNodePtr commentnode;
      int len;
      xmlChar *content;

      content = Tcl_GetStringFromObj(objv[3], &len);
      commentnode = xmlNewTextLen(content, len);
      xmlAddChild(nodePtr, commentnode);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(commentnode));
    }

    break;

  case TCLDOM_DOCUMENT_CREATECDATASECTION:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 2, objv, "token text");
      return TCL_ERROR;
    }

    if (docPtr) {
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(xmlNewDocTextLen(docPtr, content, len)));
    } else {
      xmlNodePtr textnode;
      char *content;
      int len;

      content = Tcl_GetStringFromObj(objv[3], &len);
      textnode = xmlNewTextLen(content, len);
      xmlAddChild(nodePtr, textnode);

      Tcl_SetObjResult(interp, TclDOM_CreateObjFromNode(textnode));
    }

    break;

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMNodeCommand --
 *
 *  Implements dom::libxml2::node command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMNodeCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option;
  xmlNodePtr nodePtr;
  Tcl_Obj *objPtr;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], NodeCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
    return TCL_ERROR;
  }

  switch ((enum NodeCommandMethods) method) {

  case TCLDOM_NODE_CGET:

    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 3, objv, "cget option");
      return TCL_ERROR;
    }

    if (Tcl_GetIndexFromObj(interp, objv[3], NodeCommandOptions,
			    "option", 0, &option) != TCL_OK) {
      return TCL_ERROR;
    }

    switch ((enum NodeCommandOptions) option) {

    case TCLDOM_NODE_NODETYPE:

      switch (nodePtr->type) {
      case XML_ELEMENT_NODE:
	Tcl_SetResult(interp, "element", NULL);
	break;
      case XML_ATTRIBUTE_NODE:
	Tcl_SetResult(interp, "attribute", NULL);
	break;
      case XML_TEXT_NODE:
      case XML_CDATA_SECTION_NODE:
	Tcl_SetResult(interp, "textNode", NULL);
	break;
      case XML_ENTITY_REF_NODE:
	Tcl_SetResult(interp, "entityReference", NULL);
	break;
      case XML_ENTITY_NODE:
	Tcl_SetResult(interp, "entity", NULL);
	break;
      case XML_PI_NODE:
	Tcl_SetResult(interp, "processingInstruction", NULL);
	break;
      case XML_COMMENT_NODE:
	Tcl_SetResult(interp, "comment", NULL);
	break;
      case XML_DOCUMENT_NODE:
	Tcl_SetResult(interp, "document", NULL);
	break;
      case XML_DOCUMENT_TYPE_NODE:
	Tcl_SetResult(interp, "docType", NULL);
	break;
      case XML_DOCUMENT_FRAG_NODE:
	Tcl_SetResult(interp, "documentFragment", NULL);
	break;
      case XML_NOTATION_NODE:
	Tcl_SetResult(interp, "notation", NULL);
	break;
      case XML_HTML_DOCUMENT_NODE:
	Tcl_SetResult(interp, "HTMLdocument", NULL);
	break;
      case XML_DTD_NODE:
	Tcl_SetResult(interp, "dtd", NULL);
	break;
      case XML_ELEMENT_DECL:
	Tcl_SetResult(interp, "elementDecl", NULL);
	break;
      case XML_ATTRIBUTE_DECL:
	Tcl_SetResult(interp, "attributeDecl", NULL);
	break;
      case XML_ENTITY_DECL:
	Tcl_SetResult(interp, "entityDecl", NULL);
	break;
      case XML_NAMESPACE_DECL:
	Tcl_SetResult(interp, "namespaceDecl", NULL);
	break;
      case XML_XINCLUDE_START:
	Tcl_SetResult(interp, "xincludeStart", NULL);
	break;
      case XML_XINCLUDE_END:
	Tcl_SetResult(interp, "xincludeEnd", NULL);
	break;
      default:
	Tcl_SetResult(interp, "unknown", NULL);
      }

      break;

    case TCLDOM_NODE_NODENAME:

      objPtr = Tcl_NewStringObj(nodePtr->name, -1);
      Tcl_SetObjResult(interp, objPtr);

      break;

    default:
      Tcl_SetResult(interp, "unknown option", NULL);
      return TCL_ERROR;
    }

    break;

  default:
    Tcl_SetResult(interp, "not yet implemented", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMElementCommand --
 *
 *  Implements dom::libxml2::element command.
 *
 * Results:
 *  Depends on method.
 *
 * Side effects:
 *  Depends on method.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOMElementCommand (dummy, interp, objc, objv)
     ClientData dummy;
     Tcl_Interp *interp;
     int objc;
     Tcl_Obj *CONST objv[];
{
  int method, option;
  xmlNodePtr nodePtr;
  Tcl_Obj *objPtr;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "method ?args...?");
    return TCL_ERROR;
  }

  if (Tcl_GetIndexFromObj(interp, objv[1], ElementCommandMethods,
			  "method", 0, &method) != TCL_OK) {
    return TCL_ERROR;
  }

  if (TclDOM_GetNodeFromObj(interp, objv[2], &nodePtr) != TCL_OK) {
    return TCL_ERROR;
  }

  Tcl_SetResult(interp, "element command not yet implemented", NULL);
  return TCL_ERROR;

  switch ((enum ElementCommandMethods) method) {

  case TCLDOM_ELEMENT_CGET:
    break;

  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_CreateObjFromDoc --
 *
 *  Create a Tcl_Obj to wrap a document.
 *
 * Q: What if this document is already wrapped?
 *
 * Results:
 *  Returns Tcl_Obj*.
 *
 * Side effects:
 *  Allocates object.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOM_CreateObjFromDoc (docPtr)
  xmlDocPtr docPtr;
{
  Tcl_Obj *objPtr;

  objPtr = Tcl_NewObj();
  objPtr->internalRep.otherValuePtr = (VOID *) docPtr;
  objPtr->typePtr = &TclDOM_DocObjType;
  objPtr->bytes = Tcl_Alloc(20);
  sprintf(objPtr->bytes, "doc%d", docCntr++);
  objPtr->length = strlen(objPtr->bytes);

  return objPtr;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_CreateObjFromNode --
 *
 *  Create a Tcl_Obj to wrap a tree node.
 *
 *  NB. We could get alot fancier in generating a
 *  string rep, eg. to allow mapping from a string
 *  rep back to an xmlNodePtr.  However, in this
 *  implementation the application will have to
 *  keep hold of the internal rep.
 *
 * Results:
 *  Returns Tcl_Obj*.
 *
 * Side effects:
 *  Allocates object.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOM_CreateObjFromNode (nodePtr)
  xmlNodePtr nodePtr;
{
  Tcl_Obj *objPtr;

  objPtr = Tcl_NewObj();
  objPtr->internalRep.otherValuePtr = (VOID *) nodePtr;
  objPtr->typePtr = &TclDOM_NodeObjType;
  objPtr->bytes = Tcl_Alloc(strlen("domnode") + 1);
  strcpy(objPtr->bytes, "domnode");
  objPtr->length = strlen("domnode");

  return objPtr;
}

/*
 *	Manage TclDOM objects
 */

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_GetDocFromObj --
 *
 *  Gets an xmlDocPtr from a Tcl_Obj.
 *
 * Results:
 *  Returns success code.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOM_GetDocFromObj (interp, objPtr, docPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     xmlDocPtr *docPtr;
{

  if (objPtr->typePtr == &TclDOM_DocObjType) {
    *docPtr = (xmlDocPtr) objPtr->internalRep.otherValuePtr;
  } else if (TclDOM_DocSetFromAny(interp, objPtr)) {
    *docPtr = (xmlDocPtr) objPtr->internalRep.otherValuePtr;
  } else {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOM_GetNodeFromObj --
 *
 *  Gets an xmlNodePtr from a Tcl_Obj.
 *
 * Results:
 *  Returns success code.
 *
 * Side effects:
 *  None.
 *
 *----------------------------------------------------------------------------
 */

int
TclDOM_GetNodeFromObj (interp, objPtr, nodePtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
     xmlNodePtr *nodePtr;
{

  if (objPtr->typePtr == &TclDOM_NodeObjType) {
    *nodePtr = objPtr->internalRep.otherValuePtr;
  } else if (TclDOM_NodeSetFromAny(interp, objPtr)) {
    *nodePtr = objPtr->internalRep.otherValuePtr;
  } else {
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMNewDoc --
 *
 *  Creates a new xmlDocPtr and wraps it in a Tcl_Obj.
 *
 * Results:
 *  Returns a *TclObj
 *
 * Side effects:
 *  Objects allocated.
 *
 *----------------------------------------------------------------------------
 */

Tcl_Obj *
TclDOMNewDoc(interp)
     Tcl_Interp *interp;
{
  Tcl_Obj *objPtr;
  xmlDocPtr new;

  new = xmlNewDoc("1.0");
  if (!new) {
    return NULL;
  }

  objPtr = Tcl_NewObj();

  objPtr->internalRep.otherValuePtr = (VOID *) new;
  objPtr->typePtr = &TclDOM_DocObjType;
  objPtr->bytes = Tcl_Alloc(20);
  sprintf(objPtr->bytes, "doc%d", docCntr++);
  objPtr->length = strlen(objPtr->bytes);

  return objPtr;
}

int
TclDOM_DocSetFromAny(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entryPtr;

  entryPtr = Tcl_FindHashEntry(&documents, Tcl_GetStringFromObj(objPtr, NULL));

  if (entryPtr) {
    if (objPtr->typePtr != NULL && objPtr->typePtr->freeIntRepProc != NULL) {
      objPtr->typePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = Tcl_GetHashValue(entryPtr);
    objPtr->typePtr = &TclDOM_DocObjType;
  } else {
    Tcl_SetResult(interp, "no such DOM document - try parsing first", NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

void
TclDOM_DocUpdate(objPtr)
     Tcl_Obj *objPtr;
{
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  entryPtr = Tcl_FirstHashEntry(&documents, &search);
  while (entryPtr && objPtr->internalRep.otherValuePtr != Tcl_GetHashValue(entryPtr)) {
    entryPtr = Tcl_NextHashEntry(&search);
  }
  if (entryPtr == NULL) {
    objPtr->bytes = NULL;
    objPtr->length = 0;
  } else {
    strcpy(objPtr->bytes, Tcl_GetHashKey(&documents, entryPtr));
    objPtr->length = strlen(Tcl_GetHashKey(&documents, entryPtr));
  }
}

void
TclDOM_DocDup(srcPtr, dstPtr)
     Tcl_Obj *srcPtr;
     Tcl_Obj *dstPtr;
{
  Tcl_InvalidateStringRep(dstPtr);

  dstPtr->internalRep.otherValuePtr = (VOID *) xmlCopyDoc((xmlDocPtr) srcPtr->internalRep.otherValuePtr, 1);
  dstPtr->bytes = Tcl_Alloc(20);
  sprintf(dstPtr->bytes, "doc%d", docCntr++);
  dstPtr->length = strlen(dstPtr->bytes);
}

void
TclDOM_DocFree(objPtr)
     Tcl_Obj *objPtr;
{
  xmlFreeDoc((xmlDocPtr) objPtr->internalRep.otherValuePtr);
}

int
TclDOM_NodeSetFromAny(interp, objPtr)
     Tcl_Interp *interp;
     Tcl_Obj *objPtr;
{
  Tcl_SetResult(interp, "unable to find DOM node", NULL);
  return TCL_ERROR;
}

void
TclDOM_NodeUpdate(objPtr)
     Tcl_Obj *objPtr;
{
  objPtr->bytes = Tcl_Alloc(strlen("domnode"));
  strcpy(objPtr->bytes, "domnode");
  objPtr->length = strlen("domnode");
}

/* This operation is not supported */
void
TclDOM_NodeDup(srcPtr, dstPtr)
     Tcl_Obj *srcPtr;
     Tcl_Obj *dstPtr;
{
  Tcl_InvalidateStringRep(dstPtr);
}

/* This operation does nothing - app should use dom::node interface */
void
TclDOM_NodeFree(objPtr)
     Tcl_Obj *objPtr;
{
}

/*
 *----------------------------------------------------------------------------
 *
 * TclDOMGenericError --
 *
 *  Handler for parse errors.
 *
 * Results:
 *  Stores error message.
 *
 * Side effects:
 *  Transform will return error condition.
 *
 *----------------------------------------------------------------------------
 */

static void
TclDOMGenericError (void *ctx, const char *msg, ...)
{
  va_list args;
  char buf[2048];
  int len;

  errorInfo.code = TCL_ERROR;

  if (!errorInfo.msg) {
    errorInfo.msg = Tcl_NewObj();
    Tcl_IncrRefCount(errorInfo.msg);
  }

  va_start(args,msg);
  len = vsnprintf(buf, 2047, msg, args);
  va_end(args);

  Tcl_AppendToObj(errorInfo.msg, buf, len);

}

